"""
Unit tests for connectors flow
"""

from unittest.mock import call

from asynctest import patch, CoroutineMock
import pytest

from aio_sf_streaming import (PasswordSalesforceStreaming,
                              RefreshTokenSalesforceStreaming)


@patch('aiohttp.ClientSession')
@pytest.mark.asyncio
async def test_password_flow(mock_client_session):
    """
    Test fetch token for password flow
    """
    # Mock client session mock with async context manager.
    # TODO: Found a easier way to create this
    async with mock_client_session() as sm:
        async with sm.post() as rm:
            rm.json = CoroutineMock()
            rm.json.return_value = {
                 'instance_url': 'https://foo.salesforce.com',
                 'token_type': 'Bearer',
                 'access_token': 'my_token'}
    mock_client_session.reset_mock()
    sm.post.reset_mock()

    client = PasswordSalesforceStreaming(
                username="my_username",
                password="my_password",
                client_id="my_client_id",
                client_secret="my_client_secret",
                sandbox=False)
    token, instance_url = await client.fetch_token()

    assert token == 'my_token'
    assert instance_url == 'https://foo.salesforce.com'

    assert mock_client_session.call_count == 1
    _, kwargs = mock_client_session.call_args
    assert "headers" in kwargs
    assert 'Accept' in kwargs["headers"]
    assert kwargs['headers']['Accept'] == 'application/json'

    assert sm.post.call_count == 1
    assert sm.post.call_args == call(
        'https://login.salesforce.com/services/oauth2/token',
        data={
            'grant_type': 'password',
            'username': 'my_username',
            'password': 'my_password',
            'client_id': 'my_client_id',
            'client_secret': 'my_client_secret'})


@patch('aiohttp.ClientSession')
@pytest.mark.asyncio
async def test_refresh_token_flow(mock_client_session):
    """
    Test fetch token for refresh token flow
    """
    # Mock client session mock with async context manager.
    # TODO: Found a easier way to create this
    async with mock_client_session() as sm:
        async with sm.post() as rm:
            rm.json = CoroutineMock()
            rm.json.return_value = {
                 'instance_url': 'https://foo.salesforce.com',
                 'token_type': 'Bearer',
                 'access_token': 'my_token'}
    mock_client_session.reset_mock()
    sm.post.reset_mock()

    client = RefreshTokenSalesforceStreaming(
                refresh_token="refresh_token",
                client_id="my_client_id",
                client_secret="my_client_secret",
                sandbox=False)
    token, instance_url = await client.fetch_token()

    assert token == 'my_token'
    assert instance_url == 'https://foo.salesforce.com'

    assert mock_client_session.call_count == 1
    _, kwargs = mock_client_session.call_args
    assert "headers" in kwargs
    assert 'Accept' in kwargs["headers"]
    assert kwargs['headers']['Accept'] == 'application/json'

    assert sm.post.call_count == 1
    assert sm.post.call_args == call(
        'https://login.salesforce.com/services/oauth2/token',
        data={
            'grant_type': 'refresh_token',
            'refresh_token': 'refresh_token',
            'client_id': 'my_client_id',
            'client_secret': 'my_client_secret'})


def test_missing_parameters():
    """
    Test arguments that must be provided
    """
    with pytest.raises(TypeError):
        PasswordSalesforceStreaming(
            password="my_password",
            client_id="my_client_id",
            client_secret="my_client_secret")

    with pytest.raises(TypeError):
        PasswordSalesforceStreaming(
            username="my_username",
            client_id="my_client_id",
            client_secret="my_client_secret")

    with pytest.raises(TypeError):
        PasswordSalesforceStreaming(
            username="my_username",
            password="my_password",
            client_secret="my_client_secret")

    with pytest.raises(TypeError):
        PasswordSalesforceStreaming(
            username="my_username",
            password="my_password",
            client_id="my_client_id")

    with pytest.raises(TypeError):
        RefreshTokenSalesforceStreaming(
            refresh_token="refresh_token",
            client_secret="my_client_secret")

    with pytest.raises(TypeError):
        RefreshTokenSalesforceStreaming(
            refresh_token="refresh_token",
            client_id="my_client_id")

    with pytest.raises(TypeError):
        RefreshTokenSalesforceStreaming(
            client_id="my_client_id",
            client_secret="my_client_secret")



"""
Core module: define the root base class with main logic
"""
import abc
import asyncio
import copy
import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import aiohttp

logger = logging.getLogger("aio_sf_streaming")

# Typing utils
JSONObject = Dict[str, Any]
JSONList = List[JSONObject]
JSONType = Union[JSONObject, JSONList]


class BaseSalesforceStreaming(abc.ABC):
    """
    Base low-level *aio-sf-streaming* class.

    Can not be used directly: must be sub-classed with at least one connector
    implementation. The class provide basic functionalities. Additional
    functionalities can be added with provided mixins.

    The main logic is implemented here but you should not use it directly.

    :param sandbox: If ``True``, the connexion will be made on a sandbox,
        from ``https://test.salesforce.com`` instead of the main login
        route at ``https://login.salesforce.com``.
    :param version: The API version to use. For example ``'42.0'``.
    :param loop: Asyncio loop used
    :param connector: ``aiohttp`` connector used for main session. Mainly used
        for test purpose.

    This class supports the context manager protocol for self closing.

    All main members are coroutine, even if default implementation does do any
    asynchronous call. With this convention, sub classes and mixins can easily
    override this members and do complex call.

    See :py:class:`SimpleSalesforceStreaming` for an usage example.
    """

    version: str  #: SF api version to use
    sandbox: bool  #: Use test server
    connector: aiohttp.BaseConnector  #: aiohttp connector for main session
    instance_url: Optional[str]  #: Instance url (retrieved with token)
    session: Optional[aiohttp.ClientSession]  #: Underlying connection
    client_id: Optional[str]  #: The client id token from handshake
    message_count: int  #: Message id count
    timeout: int  #: Timeout connection duration
    should_stop: bool  #: Set to True to stop streaming

    #: Header used in all requests
    base_header: dict = {"Accept": "application/json"}

    def __init__(
        self,
        *,
        sandbox: bool = False,
        version: str = "42.0",
        loop: asyncio.AbstractEventLoop = None,
        connector: aiohttp.BaseConnector = None,
    ) -> None:
        self.version = version
        self.sandbox = sandbox
        self._loop = loop
        self.connector = connector
        self.instance_url = None
        self.session = None
        self.client_id = None
        self.message_count = 0
        self.timeout = 120
        self.should_stop = False
        super().__init__()

    # -------------------- High level api --------------------

    async def start(self) -> None:
        """
        Connect to Salesforce, authenticate and init CometD connexion.

        A best practice is to use async context manager interface that will
        call this method directly.
        """
        self.session = await self.create_connected_session()
        await self.handshake()

    async def subscribe(self, channel: str) -> JSONList:
        """
        Subscribe to a channel. Can be used directly::

            await client.subscribe('/topic/Foo')

        This method, and the underlying protocol, are safe to be started as
        an background task::

            loop.create_task(client.subscribe('/topic/Foo'))

        """
        response = await self.send(await self.get_subscribe_payload(channel))
        logger.info("Subscribe response: %r", response)
        return response

    async def messages(self) -> JSONObject:
        """
        Asynchronous generator that fetch new messages and return one as soon
        as one is available::

            async for message in client.messages():
                channel = message['channel']
                print(channel, ':', message)

        This method iterate over **all** messages, even on internal/meta one.
        If you want to only iterate over messages from channels you subscribed,
        you should use :py:func:`BaseSalesforceStreaming.events`.

        .. warning::
            Linked to the underlying protocol, long-pooling based, the client
            should reconnect as soon as possible. Practically, client have 40
            seconds to reconnect. If your processing take a longer time, a new
            connection should be made. You should avoid doing long processing
            between each iteration or launch this processing into a background
            task.

        """
        while True:
            if self.should_stop:
                return
            try:
                response = await self.send(
                    {"channel": "/meta/connect", "connectionType": "long-polling"}
                )
            except asyncio.TimeoutError:
                logger.info("Timeout")
                continue
            except aiohttp.ClientResponseError as error:
                if error.code == 408:
                    # Timeout
                    logger.info("Timeout")
                    continue
                else:
                    raise

            if self.should_stop:
                return

            logger.debug("Messages: received %r", response)
            for message in response:
                if self.should_stop:
                    return
                yield message

    async def events(self) -> JSONObject:
        """
        Asynchronous generator that fetch new events and return one as soon
        as one is available::

            async for message in client.events():
                channel = message['channel']
                print(channel, ':', message)

        This method is different from :py:func:`BaseSalesforceStreaming.messages`
        because it filter messages and provide only those related to the
        channels you subscribed.
        """
        async for message in self.messages():
            if not message.get("channel", "").startswith("/meta/"):
                yield message

    async def ask_stop(self) -> None:
        """
        Ask client to stop receiving event::

            async for event in client.events():
                ...
                if ...:
                    await client.ask_stop()

        This call will eventually stop
        :py:func:`BaseSalesforceStreaming.messages` and
        :py:func:`BaseSalesforceStreaming.events` async generator but this can
        take some time if not called inside the loop body: the generator will
        wait a timeout response from Salesforce server.
        """
        self.should_stop = True

    async def unsubscribe(self, channel: str) -> JSONList:
        """
        Unsubscribe to a channel. Can be used directly::

            await client.unsubscribe('/topic/Foo')

        This method, and the underlying protocol, are safe to be started as
        an background task::

            loop.create_task(client.unsubscribe('/topic/Foo'))

        """
        response = await self.send(await self.get_unsubscribe_payload(channel))
        logger.info("Unsubscribe response: %r", response)
        return response

    async def stop(self) -> None:
        """
        Disconnect to Salesforce and close underlying connection.

        A best practice is to use async context manager interface that will
        call this method directly.
        """
        await self.ask_stop()
        await self.disconnect()
        await self.close_session()

    # -------------------- Connection logic --------------------

    @property
    def token_url(self) -> str:
        """
        The url that should be used to fetch an access token.
        """
        url_prefix = "test" if self.sandbox else "login"
        return f"https://{url_prefix}.salesforce.com/services/oauth2/token"

    @abc.abstractmethod
    async def fetch_token(self) -> Tuple[str, str]:
        """
        Abstract coroutine method of connector that must provide an access
        token and the instance url linked.
        """

    async def create_connected_session(self) -> aiohttp.ClientSession:
        """
        This coroutine create an ``aiohttp.ClientSession`` using fetched token
        """
        token, self.instance_url = await self.fetch_token()

        base_header = copy.deepcopy(self.base_header)
        base_header.update({"Authorization": f"Bearer {token}"})

        session = aiohttp.ClientSession(
            connector=self.connector, headers=base_header, loop=self.loop
        )
        return session

    async def close_session(self) -> None:
        """
        Close the underlying ``aiohttp.ClientSession`` connection
        """
        if self.session is None:
            return
        await self.session.close()
        self.session = None

    # -------------------- Bayeux/CometD logic layer --------------------

    @property
    def end_point(self) -> str:
        """
        Cometd endpoint
        """
        return f"/cometd/{self.version}/"

    async def get_handshake_payload(self) -> JSONObject:
        """
        Provide the handshake payload
        """
        return {
            "channel": "/meta/handshake",
            "supportedConnectionTypes": ["long-polling"],
            "version": "1.0",
            "minimumVersion": "1.0",
        }

    async def get_subscribe_payload(self, channel: str) -> JSONObject:
        """
        Provide the subscription payload for a specific channel
        """
        return {"channel": "/meta/subscribe", "subscription": channel}

    async def get_unsubscribe_payload(self, channel: str) -> JSONObject:
        """
        Provide the unsubscription payload for a specific channel
        """
        return {"channel": "/meta/unsubscribe", "subscription": channel}

    async def send(self, data: JSONObject) -> JSONType:
        """
        Send data to CometD server when the connection is established::

            # Manually disconnect
            await client.send({'channel': '/meta/disconnect'})

        """
        self.message_count += 1

        # Add  id and client_id to payload
        data = copy.copy(data)
        data["id"] = str(self.message_count)
        if self.client_id:
            data["clientId"] = self.client_id

        # Post data
        return await self.post(self.end_point, json=data)

    async def handshake(self) -> JSONList:
        """
        Coroutine that perform an handshake (mandatory before any other action)
        """
        self.message_count = 0

        response = await self.send(await self.get_handshake_payload())
        logger.info("Handshake response: %r", response)
        self.client_id = response[0]["clientId"]

        return response

    async def disconnect(self) -> JSONList:
        """
        Disconnect from the SF streaming server
        """
        return await self.send({"channel": "/meta/disconnect"})

    # -------------------- IO layer helpers --------------------

    async def get(self, sub_url: str, **kwargs) -> JSONType:
        """
        Perform a simple json get request from an internal url::

            response = await.client.get('/myendpoint/')

        """
        return await self.request("get", sub_url, **kwargs)

    async def post(self, sub_url: str, **kwargs) -> JSONType:
        """
        Perform a simple json post request from an internal url::

            response = await.client.post('/myendpoint/', json={'data': 'foo'})

        """
        return await self.request("post", sub_url, **kwargs)

    async def request(self, method: str, sub_url: str, **kwargs) -> JSONType:
        """
        Perform a simple json request from an internal url
        """
        url = f"{self.instance_url}{sub_url}"
        logger.debug("Perform %r to %r with %r", method, url, kwargs)

        async with self.session.request(
            method, url, timeout=self.timeout, **kwargs
        ) as resp:
            resp.raise_for_status()
            data = await resp.json()

        return data

    # -------------------- SPECIALS METHODS -------------------- #

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """
        Running event loop
        """
        if self._loop is None:
            self._loop = asyncio.get_event_loop()
        return self._loop

    # Asynchronous context manager

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()



"""
Connectors module: Provide authentication implementation
"""

from typing import Tuple

import aiohttp

from .core import BaseSalesforceStreaming


class BaseConnector(BaseSalesforceStreaming):
    """
    Base class for all sf connectors.

    :param client_id: OAuth2 client Id (mandatory)
    :param client_secret: Oauth2 client secret (mandatory)
    :param login_connector: aiohttp connector used during connection.
        Mainly used for test purpose.

    See :class:`.BaseSalesforceStreaming` for other keywords arguments.
    """

    def __init__(
        self,
        *,
        client_id: str = None,
        client_secret: str = None,
        login_connector: aiohttp.BaseConnector = None,
        **kwargs
    ):
        if any(v is None for v in (client_id, client_secret)):
            raise TypeError("All credentials arguments are mandatory")

        self.login_connector = login_connector
        self.credentials = {"client_id": client_id, "client_secret": client_secret}
        super().__init__(**kwargs)

    async def fetch_token(self) -> Tuple[str, str]:
        # use a temporary session only to fetch token because client session
        # does not seems to allow update default headers on a already created
        # session
        async with aiohttp.ClientSession(
            connector=self.login_connector, headers=self.base_header, loop=self.loop
        ) as session:
            async with session.post(self.token_url, data=self.credentials) as resp:
                data = await resp.json()

        assert data["token_type"] == "Bearer"
        instance_url = data["instance_url"]
        access_token = data["access_token"]

        return access_token, instance_url


class PasswordSalesforceStreaming(BaseConnector):
    """
    Create a SF streaming manager with password flow connection.

    Main arguments are connection credentials:

    :param username: User login name
    :param password: User password

    See :class:`.BaseConnector` for other keywords arguments.
    """

    def __init__(self, *, username: str = None, password: str = None, **kwargs):
        if any(v is None for v in (username, password)):
            raise TypeError("All credentials arguments are mandatory")

        super().__init__(**kwargs)
        # Credentials used to fetch access token
        self.credentials.update(
            {"grant_type": "password", "username": username, "password": password}
        )


class RefreshTokenSalesforceStreaming(BaseConnector):
    """
    Create a SF streaming manager with password refresh token connection.

    Main arguments are connection credentials:

    :param refresh_token: Refresh token

    See :class:`.BaseConnector` for other keywords arguments.
    """

    def __init__(self, *, refresh_token: str = None, **kwargs):
        if refresh_token is None:
            raise TypeError("All credentials arguments are mandatory")

        super().__init__(**kwargs)
        # Credentials used to fetch access token
        self.credentials.update(
            {"grant_type": "refresh_token", "refresh_token": refresh_token}
        )



"""
Mixins module: Provide various mixins modules
"""
import asyncio
import enum
import logging
from typing import Union

from .core import JSONList, JSONObject

logger = logging.getLogger("aio_sf_streaming")


class TimeoutAdviceMixin:
    """
    Simple mixin that automatically set timeout setting according to SF
    advice, if provided.
    """

    async def messages(self) -> JSONObject:
        """
        See :py:func:`BaseSalesforceStreaming.messages`
        """
        async for message in super().messages():
            if message.get("channel", "") == "/meta/connect" and "advice" in message:
                timeout_advice = message["advice"].get("timeout", None)
                if timeout_advice:
                    self.timeout = timeout_advice / 1000
            yield message


class ReplayType(enum.Enum):
    """
    Enumeration with special replay values
    """

    ALL_EVENTS = -2  #: Replay all events available.
    NEW_EVENTS = -1  #: No replay, retrieve only new events.


class ReplayMixin:
    """
    Mixing adding replay support to the streaming client.

    This mixin is not enough, you must implement :py:func:`ReplayMixin.store_replay_id` and
    `:py:func:`ReplayMixin.get_last_replay_id` in a subclass in order to have a working replay.
    """

    async def get_handshake_payload(self) -> JSONObject:
        """
        See :py:func:`BaseSalesforceStreaming.get_handshake_payload`
        """
        payload = await super().get_handshake_payload()
        # Activate replay extension
        payload.setdefault("ext", {}).update({"replay": True})
        return payload

    async def get_subscribe_payload(self, channel: str) -> JSONObject:
        """
        See :py:func:`BaseSalesforceStreaming.get_subscribe_payload`
        """
        payload = await super().get_subscribe_payload(channel)

        # Call inner callback to retrieve the last replay id
        replay_id = await self.get_last_replay_id(channel)

        # No response => Use only new events (default behavior)
        if not replay_id:
            replay_id = ReplayType.NEW_EVENTS

        # Extract replay value
        if isinstance(replay_id, ReplayType):
            replay_id = replay_id.value
        replay_id = int(replay_id)

        # Update payload
        payload.setdefault("ext", {}).setdefault("replay", {})
        payload["ext"]["replay"][channel] = replay_id

        return payload

    async def messages(self) -> JSONObject:
        """
        See :py:func:`BaseSalesforceStreaming.messages`
        """
        async for message in super().messages():
            channel = message["channel"]

            # On new message, call callback to store replay id
            if not channel.startswith("/meta/"):
                event = message["data"]["event"]
                replay_id = event["replayId"]
                creation_time = event["createdDate"]

                # Create a task : do not wait the replay id is stored to
                # reconnect as soon as possible
                self.loop.create_task(self.store_replay_id(channel, replay_id, creation_time))
            yield message

    async def store_replay_id(self, channel: str, replay_id: int, creation_time: str) -> None:
        """
        Callback called to store a replay id. You should override this method
        to implement your custom logic.

        :param channel: Channel name
        :param replay_id: replay id to store
        :param creation_time: Creation time. You should store only the last
            created object but you can not know if you received event in order
            without this. This value is the string provided by SF.
        """

    async def get_last_replay_id(self, channel: str) -> Union[ReplayType, int]:
        """
        Callback called to retrieve a replay id. You should override this method
        to implement your custom logic.

        :param channel: Channel name
        """


class AutoVersionMixin:
    """
    Simple mixin that fetch last api version before connect.
    """

    async def handshake(self) -> JSONList:
        """
        See :py:func:`BaseSalesforceStreaming.handshake`
        """
        # Get last api version
        data = await self.get("/services/data/")
        try:
            self.version = data[-1]["version"]
        except (IndexError, KeyError):
            pass
        logger.info("API version used: %r", self.version)

        return await super().handshake()


class AutoReconnectMixin:
    """
    Mixin that will automatically reconnect when asked by Salesforce
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Used to store all subscribed channels
        self._subchannels = None

    async def start(self) -> None:
        """
        See :py:func:`BaseSalesforceStreaming.start`
        """
        self._subchannels = set()
        await super().start()

    async def subscribe(self, channel: str) -> JSONList:
        """
        See :py:func:`BaseSalesforceStreaming.subscribe`
        """
        self._subchannels.add(channel)
        return await super().subscribe(channel)

    async def messages(self) -> JSONObject:
        """
        See :py:func:`BaseSalesforceStreaming.messages`
        """
        async for message in super().messages():
            channel = message["channel"]

            # If asked, perform a new handshake
            if channel.startswith("/meta/") and message.get("error") == "403::Unknown client":
                logger.info("Disconnected, do new handshake")
                await self.handshake()
                continue

            yield message

    async def unsubscribe(self, channel: str) -> JSONList:
        """
        See :py:func:`BaseSalesforceStreaming.unsubscribe`
        """
        self._subchannels.remove(channel)
        return await super().unsubscribe(channel)

    async def stop(self) -> None:
        """
        See :py:func:`BaseSalesforceStreaming.stop`
        """
        await super().stop()
        self._subchannels = None

    async def handshake(self) -> JSONList:
        """
        See :py:func:`BaseSalesforceStreaming.handshake`
        """
        response = await super().handshake()

        # If we reconnect, we must re-subscribe to all channels
        for channel in self._subchannels:
            self.loop.create_task(super().subscribe(channel))

        return response


class ReSubscribeMixin:
    """
    Mixin that handle subscription error, will try again after a short delay

    :param retry_sub_duration: Duration between subscribe retry if server is
        too buzy (initial value).
    :param retry_factor: Factor amplification between each successive retry
    :param retry_max_duration: Maximum value of the retry duration
    :param retry_max_count: Maximum count of retry, after this count is reach,
        response or exception are propagated.
    """

    def __init__(
        self,
        retry_sub_duration: float = 0.1,
        retry_factor: float = 1.,
        retry_max_duration: float = 30.,
        retry_max_count: int = 20,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.retry_sub_duration = retry_sub_duration
        self.retry_factor = retry_factor
        self.retry_max_duration = retry_max_duration
        self.retry_max_count = retry_max_count
        self.retry_current_duration = {}
        self.retry_current_count = {}

    async def should_retry_on_exception(self, channel: str, exception: Exception) -> bool:
        """
        Callback called to process an exception raised during subscription.
        Return a boolean if we must retry. If ``False`` is returned, the exception will be
        propagated to caller.

        By-default, do return always ``False``.

        :param channel: Channel name
        :param exception: The exception raised
        """
        return False

    async def should_retry_on_error_response(self, channel: str, response: JSONObject) -> bool:
        """
        Callback called to process a response with and error message.
        Return a boolean if we must retry. If ``False`` is returned, the response will be
        returned to caller.

        By-default, retry on known 'server unavailable' response.

        :param channel: Channel name
        :param response: The response received
        """
        failure = (
            response[0]
            .get("ext", {})
            .get("sfdc", {})
            .get("failureReason", "")
        )
        return (failure.startswith("SERVER_UNAVAILABLE")
                or failure.startswith("503::Server is too busy"))

    def _update_retry_count(self, channel: str) -> bool:
        """
        Update retry count for the channel. Return a boolean if we should retry
        """
        self.retry_current_count[channel] = self.retry_current_count.get(channel, 0) + 1
        if self.retry_current_count[channel] >= self.retry_max_count:
            return False
        duration = self.retry_current_duration.get(channel, -1)
        if duration < 0:
            duration = self.retry_sub_duration
        else:
            duration = min(duration * self.retry_factor, self.retry_max_count)
        self.retry_current_duration[channel] = duration
        return True

    async def subscribe(self, channel: str) -> JSONList:
        """
        See :py:func:`BaseSalesforceStreaming.subscribe`
        """
        while True:
            try:
                response = await super().subscribe(channel)
            except Exception as e:
                should_retry = await self.should_retry_on_exception(channel, e)
                if should_retry:
                    should_retry = self._update_retry_count(channel)
                if not should_retry:
                    raise
            else:
                if response and response[0]["successful"]:
                    should_retry = False
                else:
                    should_retry = await self.should_retry_on_error_response(channel, response)
                    if should_retry:
                        should_retry = self._update_retry_count(channel)

            if not should_retry:
                self.retry_current_duration[channel] = -1
                self.retry_current_count[channel] = 0
                return response

            await asyncio.sleep(self.retry_current_duration[channel])


class AllMixin(
    TimeoutAdviceMixin,  # Use SF timeout advice
    AutoVersionMixin,  # Auto-fetch last api version
    ReplayMixin,  # Add replay support
    AutoReconnectMixin,  # Add auto-reconnection feature
    ReSubscribeMixin,
):  # Handle subscription errors
    """
    Helper class to add all mixin with one class
    """
